<?php

namespace Drp\SimpleCms\Test\Unit;

use Drp\SimpleCms\Contracts\Publishable;
use Drp\SimpleCms\Events\ContentWasPublished;
use Drp\SimpleCms\Exceptions\CouldNotResetDraft;
use Drp\SimpleCms\Exceptions\InvalidContentStatus;
use Drp\SimpleCms\Scopes\PublishedScope;
use Drp\SimpleCms\Test\Support\Fakes\TestContent;
use Drp\SimpleCms\Test\TestCase;

/**
 * Class ContentTest
 * @package Drp\SimpleCms\Test\Unit
 */
class ContentTest extends TestCase
{


    /**
     * @test
     */
    public function it_is_generated_with_a_content_id()
    {
        $content = TestContent::create();
        $content2 = TestContent::create();

        $this->assertEquals(1, $content->content_id);
        $this->assertEquals(2, $content2->content_id);
    }

    /**
     * @test
     */
    public function it_cant_have_duplicate_content_id_and_status()
    {
        $this->expectException(InvalidContentStatus::class);

        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create(['content_id' => 1]);
    }

    /**
     * @test
     */
    public function I_can_update_the_review_status()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $content->updateReviewStatus();

        $this->assertEquals(Publishable::STATUS_REVIEW, $content->status);
        $this->assertEquals(Publishable::STATUS_PUBLISHED, $publishedContent->status);

        $content->updateReviewStatus();
        $this->assertEquals(Publishable::STATUS_DRAFT, $content->status);
    }

    /**
     * @test
     */
    public function it_returns_if_it_can_be_reviewed()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertTrue($content->canBeReviewed());

        $content->updateReviewStatus();
        $this->assertTrue($content->canBeReviewed());

        $content->updatePublishStatus();
        $this->assertFalse($content->canBeReviewed());
    }

    /**
     * @test
     */
    public function it_returns_if_it_can_be_published()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)->make();

        $this->assertFalse($content->canBePublished());

        $content->save();
        $this->assertTrue($content->canBePublished());
    }

    /**
     * @test
     */
    public function it_returns_its_draft_or_review_version()
    {
        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertSameModel($draftContent, $draftContent->getDraftOrReviewVersion());

        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create(['content_id' => 1]);

        $this->assertSameModel($draftContent, $publishedContent->getDraftOrReviewVersion());

        $draftContent->updateReviewStatus();
        $this->assertSameModel($draftContent, $publishedContent->getDraftOrReviewVersion());
    }

    /**
     * @test
     */
    public function I_can_update_the_published_status()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertTrue($content->hasDraftOrReviewVersion());

        $content->updatePublishStatus();

        $this->assertCount(1, TestContent::get());
        $this->assertEquals(Publishable::STATUS_PUBLISHED, $content->status);
        $this->assertFalse($content->hasDraftOrReviewVersion());

        $content->updatePublishStatus();

        $this->assertEquals(Publishable::STATUS_UNPUBLISHED, $content->status);

        $content->updatePublishStatus();

        $this->assertEquals(Publishable::STATUS_PUBLISHED, $content->status);
    }

    /**
     * @test
     */
    public function I_can_publish_a_draft_version_and_override_a_published_version()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $draftContent->updatePublishStatus();

        $this->assertCount(1, TestContent::get());
        $this->assertNull(TestContent::find($publishedContent->id));
    }

    /**
     * @test
     */
    public function I_can_publish_a_draft_version_and_override_an_unpublished_version()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_UNPUBLISHED)
            ->create();

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $draftContent->updatePublishStatus();

        $this->assertCount(1, TestContent::get());
        $this->assertNull(TestContent::find($publishedContent->id));
    }

    /**
     * @test
     */
    public function it_returns_if_it_has_a_draft_or_review_version()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertTrue($content->hasDraftOrReviewVersion());

        $content->updatePublishStatus();
        $this->assertFalse($content->hasDraftOrReviewVersion());

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $this->assertTrue($content->hasDraftOrReviewVersion());

        $content->updateReviewStatus();
        $this->assertTrue($content->hasDraftOrReviewVersion());
    }


    /**
     * @test
     */
    public function it_returns_its_published_version()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $this->assertSameModel($publishedContent, $publishedContent->getPublishedVersion());

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $this->assertSameModel($publishedContent, $draftContent->getPublishedVersion());

        $publishedContent->updatePublishStatus(); //Set to un-published
        $this->assertNull($draftContent->getPublishedVersion());
    }

    /**
     * @test
     */
    public function it_returns_its_unpublished_version()
    {
        $unpublishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $unpublishedContent->updatePublishStatus();

        $this->assertSameModel($unpublishedContent, $unpublishedContent->getUnPublishedVersion());

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);
        $this->assertSameModel($unpublishedContent, $draftContent->getUnPublishedVersion());

        $unpublishedContent->updatePublishStatus();//Set to published
        $this->assertNull($unpublishedContent->getUnPublishedVersion());
    }

    /**
     * @test
     */
    public function it_returns_if_it_has_a_published_version()
    {
        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertFalse($draftContent->hasPublishedVersion());

        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create(['content_id' => 1]);

        $this->assertEquals($publishedContent->id, $draftContent->getPublishedVersion()->id);
        $this->assertTrue($draftContent->hasPublishedVersion());
    }

    /**
     * @test
     */
    public function it_returns_if_it_has_an_unpublished_version()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $this->assertFalse($content->hasUnPublishedVersion());

        $content->updatePublishStatus();
        $this->assertTrue($content->hasUnPublishedVersion());
    }

    /**
     * @test
     */
    public function the_published_event_is_fired_on_publish()
    {
        $this->expectsEvents(ContentWasPublished::class);

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $content->updatePublishStatus();
    }

    /**
     * @test
     */
    public function the_latest_version_is_returned_by_id()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $this->assertSameModel($content, TestContent::getLatestVersionById($content->content_id));

        $latest = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        //Check the published is
        $this->assertSameModel($latest, TestContent::getLatestVersionById($content->content_id));

        $latest->updateReviewStatus();
        $this->assertSameModel($latest, TestContent::getLatestVersionById($content->content_id));

    }

    /** @test */
    public function it_can_determine_if_viewing_draft_content()
    {
        $contentA = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)->make();

        $contentB = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)->make();

        $contentC = factory(TestContent::class)
            ->states(Publishable::STATUS_REVIEW)->make();

        $this->assertFalse($contentA->isDraftOrInReview());
        $this->assertTrue($contentB->isDraftOrInReview());
        $this->assertTrue($contentC->isDraftOrInReview());
    }


    /**
     * @test
     */
    public function it_returns_if_it_can_be_reset()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)->make();

        $this->assertFalse($content->canBeReset());

        $content->save();
        $this->assertFalse($content->canBeReset());

        factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);
        $this->assertTrue($content->canBeReset());
    }

    /**
     * @test
     */
    public function it_can_reset_draft()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $draftContent->resetDraft();
        $this->assertFalse($publishedContent->hasDraftOrReviewVersion());
    }

    /**
     * @test
     */
    public function it_cant_reset_draft_if_there_is_no_published_version()
    {
        $this->expectException(CouldNotResetDraft::class);

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $draftContent->resetDraft();

    }

    /**
     * @test
     */
    public function it_cant_reset_draft_if_there_is_no_draft_version()
    {
        $this->expectException(CouldNotResetDraft::class);

        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $publishedContent->resetDraft();
    }


    /** @test */
    public function it_can_determine_if_it_is_a_none_published_version()
    {
        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $this->assertTrue($content->isDraftOrInReview());
    }

    /**
     * @test
     */
    public function it_returns_all_its_latest_studies()
    {
        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $result = TestContent::latestContent();

        $this->assertCount(2, $result);
        $this->assertFalse($result->contains(function ($content) {
            //Should have returned the Draft version of the first content
            return $content->id == 1 && $content->status === Publishable::STATUS_PUBLISHED;
        }));
    }

    /**
     * @test
     */
    public function it_scopes_the_draft_status()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $this->assertSameModel($content, TestContent::whereDraft()->first());
    }

    /**
     * @test
     */
    public function it_scopes_the_review_status()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_REVIEW)
            ->create();

        $this->assertSameModel($content, TestContent::whereinReview()->first());
    }


    /**
     * @test
     */
    public function the_published_or_first_draft_can_be_retrieved_for_the_admin_listing()
    {
        $publishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $awaitingContent = factory(TestContent::class)
            ->states(Publishable::STATUS_REVIEW)
            ->create();

        $unpublishedContent = factory(TestContent::class)
            ->states(Publishable::STATUS_UNPUBLISHED)
            ->create();

        $results = TestContent::getAllVersionsPrioritisedByStatus();

        $this->assertCount(4, $results);
        $this->assertSameModel($publishedContent, $results[0]);
        $this->assertSameModel($draftContent, $results[1]);
        $this->assertSameModel($awaitingContent, $results[2]);
        $this->assertSameModel($unpublishedContent, $results[3]);

    }

    /**
     * @test
     */
    public function it_returns_and_array_of_content_for_the_admin_realtionship_select()
    {
        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create(
                [
                    'title' => 'Study title'
                ]
            );

        factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $results = TestContent::optionsForSelect();

        $this->assertCount(2, $results);

        $this->assertEquals('Study title', array_first($results));
        $this->assertEquals(1, array_first(array_keys($results)));
    }

    /**
     * @test
     */
    public function the_content_can_be_deleted()
    {
        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $content->destroyAllVersion();

        $this->dontSeeInDatabase('test_contents', [
            'content_id' => 1
        ]);
        $this->assertEmpty(TestContent::all());
    }

    /**
     * @test
     */
    public function it_returns_its_relationship_name()
    {
        $content = new TestContent();
        $this->assertEquals('testContent', $content->relationshipName());
    }

    /**
     * @test
     */
    public function it_returns_its_route_parameters()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $this->assertEquals([
            'type' => 'test-content',
            'content_id' => 1
        ], $content->getContentRouteParameters());
    }

    /**
     * @test
     */
    public function it_returns_its_version_route_parameters()
    {
        factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $this->assertEquals([
            'type' => 'test-content',
            'content_id' => 2
        ], $content->getVersionRouteParameters());
    }

    /**
     * @test
     */
    public function it_returns_its_admin_url()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $this->assertEquals('http://localhost/admin/content/test-content/manage/1', $content->adminEditUrl());
    }


    /** @test */
    public function the_publish_scope_can_be_applied_content()
    {
        $content = factory(TestContent::class)
            ->states(Publishable::STATUS_PUBLISHED)
            ->create();

        $draftContent = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create(['content_id' => 1]);

        $draftContent2 = factory(TestContent::class)
            ->states(Publishable::STATUS_DRAFT)
            ->create();

        $results = TestContent::all();

        $this->assertCount(3, $results);

        TestContent::addGlobalScope(new PublishedScope());

        $results = TestContent::all();

        $this->assertCount(1, $results);
        $this->assertContainsModel($content, $results);
        $this->assertNotContainsModel($draftContent, $results);
        $this->assertNotContainsModel($draftContent2, $results);
    }
}
